use std::net::IpAddr;

use rocket::{
	form::{Context, Contextual, Error, Form},
	http::CookieJar,
	response::Redirect,
	serde::Serialize,
};
use rocket_dyn_templates::{context, Template};
use sqlx::Acquire;

use crate::{
	auth::session::Admin,
	db::{admin, doll, schema::DollTagsDb, user},
	ids::{id_public_to_db, PublicId},
	pages::{pop_toast, write_toast, CommonTemplateState},
};

use super::error_handlers::PageResult;

#[get("/")]
pub async fn index(
	meta: CommonTemplateState,
	mut db: DollTagsDb,
	user: Admin,
	jar: &'_ CookieJar<'_>,
) -> PageResult {
	let mut trx = db.begin().await?;
	let service_status = admin::get_service_status(&mut trx).await?;
	trx.commit().await?;

	let toast = pop_toast(jar);

	Ok(Template::render(
		"admin/index",
		context! {
			meta: meta.for_user(&user.0),
			service_status,
			previous: context! {
				tag_handover: Context::default(),
			},
			toast,
		},
	)
	.into())
}

#[derive(Debug, FromFormField, Serialize)]
#[serde(crate = "rocket::serde")]
pub enum SelectedForm {
	TagHandover,
}

#[derive(Debug, FromForm)]
pub struct Forms<'a> {
	pub form: SelectedForm,
	pub tag_handover: Contextual<'a, TagHandover<'a>>,
}

/// extracts one of the forms' contexts from the form group,
/// otherwise uses the $def value (usually you'll give [`&Context::default()`][`Context::default()`])
macro_rules! get_for_form {
	($def:expr, $curr:expr, $target:pat, $value:expr) => {
		if matches!($curr, $target) {
			$value
		} else {
			$def
		}
	};
}

#[derive(Debug, FromForm, Clone)]
pub struct TagHandover<'a> {
	#[field(validate=crate::ids::validate_id())]
	pub tag_id: &'a str,
	#[field(validate=len(..=256))]
	pub dest_account: &'a str,
}

#[post("/", data = "<form>")]
pub async fn handle_in_page_forms(
	meta: CommonTemplateState,
	mut db: DollTagsDb,
	user: Admin,
	mut form: Form<Forms<'_>>,
	client_ip: IpAddr,
) -> PageResult {
	match form.form {
		SelectedForm::TagHandover => {
			if let Some(values) = &form.tag_handover.value.clone() {
				let target_user = user::get(&mut *db, values.dest_account).await?;
				let user_valid = match target_user {
					Some(user) => {
						if !user.enabled {
							form.tag_handover.context.push_error(
								Error::validation("this user's account is deactivated")
									.with_name("tag_handover.dest_account"),
							);
							false
						} else {
							true
						}
					}
					None => {
						form.tag_handover.context.push_error(
							Error::validation("this user doesn't exist")
								.with_name("tag_handover.dest_account"),
						);
						false
					}
				};

				let target_tag = doll::get(
					&mut *db,
					id_public_to_db(values.tag_id)
						.expect("is form-validated so should always succeed"),
					"",
					true,
				)
				.await?;
				if target_tag.is_none() {
					form.tag_handover.context.push_error(
						Error::validation("no tag exists with this ID")
							.with_name("tag_handover.tag_id"),
					);
				}

				if user_valid && target_tag.is_some() {
					let tag = target_tag.unwrap();
					warn!(
						"[audit|{}] [{}] beginning tag handover of tag {} to \"{}\" (previous user id: {})",
						client_ip,
						user.0.id.to_string(),
						tag.id,
						values.dest_account,
						tag.bound_to_id,
					);
					return Ok(Redirect::to(uri!(
						"/admin",
						show_confirm_tag_handover(PublicId(tag.id), values.dest_account)
					))
					.into());
				}
			}
		}
	};

	let mut trx = db.begin().await?;
	let service_status = admin::get_service_status(&mut trx).await?;
	trx.commit().await?;

	let def = Context::default();
	Ok(Template::render(
		"admin/index",
		context! {
			meta: meta.for_user(&user.0),
			service_status,
			previous: context! {
				tag_handover: get_for_form!(
					&def,
					&form.form,
					SelectedForm::TagHandover,
					&form.tag_handover.context
				),
			},
		},
	)
	.into())
}

#[get("/tag-handover/<id>/<dest_username>")]
pub fn show_confirm_tag_handover(
	meta: CommonTemplateState,
	user: Admin,
	id: PublicId,
	dest_username: &str,
) -> Template {
	let tag_id = id.0;

	Template::render(
		"admin/confirm_tag_handover",
		context! {
			meta: meta.for_user(&user.0),
			tag_id,
			dest_username,
			proceed_url: uri!("/admin", handle_tag_handover(id, dest_username)),
		},
	)
}

#[get("/YES-GIMME/<id>/<dest_username>")]
pub async fn handle_tag_handover<'a>(
	mut db: DollTagsDb,
	user: Admin,
	client_ip: IpAddr,
	jar: &'a CookieJar<'_>,
	id: PublicId,
	dest_username: &'a str,
) -> PageResult {
	// note: there is currently no trace of the previous user in this handover code's audit log.
	let id = id.0;
	match user::get(&mut *db, dest_username).await? {
		Some(u) => {
			admin::handover_tag(&mut *db, id, &u.id).await?;
			warn!(
				"[audit|{}] [{}] handed over tag {} to \"{}\"",
				client_ip,
				user.0.id.to_string(),
				id,
				dest_username
			);
			write_toast(
				jar,
				format!("tag successfully handed over to \"{}\"", dest_username),
			);
		}
		None => {
			warn!(
				"[audit|{}] [{}] tried to hand over tag {} to nonexistent user \"{}\"",
				client_ip,
				user.0.id.to_string(),
				id,
				dest_username
			);
			write_toast(jar, String::from("this username doesn't exist"));
		}
	};

	Ok(Redirect::to("/admin").into())
}
